;; Shortest path examples using reduction, spreading, collision-free
;; communication.  Each processor plays three roles:
;;    1) Edge from its node to another, in which it calculates
;;       length of path to the other
;;    2) Recipient of path length from a processor as in 1, in which
;;       it determines whether it represents a minimal path (so far)
;;    3) Representative of its node, in which it determines whether
;;       a minimal path has been found yet (all participate)
;;  This is an experimental version.  Modifications are:
;;     - Separate Boolean to mark active senders
;;     - Calculates `eps' over edges leaving current candidates
;;       (each cycle) rather than single, global calculation

(in-package '*lisp)

(defconstant INF 99999999)    ;; Sufficiently large but representable

(*proclaim '(type fixnum mxn mxe))

(defvar mxn 0)   ;; Maximum number of nodes = grid size in 0 coord
(defvar mxe 0)   ;; Maximum degree of nodes = grid size in 1 coord


(*proclaim '(type (signed-pvar 32) ptd myd lng fr))

(*defvar ptd (!! INF))       ;; Minimum distance (so far) to node, or INF
(*defvar myd (!! INF))       ;; Distance arriving through this rcvr, or INF
(*defvar lng (!! 0))         ;; Length of edge leaving my node
(*defvar  fr (!! -1))        ;; Node from which I receive, or -1


(*proclaim '(type (unsigned-pvar *current-send-address-length*) to))

(*defvar to (!! 0))          ;; Pointer to receiver


(*proclaim '(type (unsigned-pvar 20) myx myy))

(*defvar myx (self-address-grid!! (!! 0)))  ;; My node number
(*defvar myy (self-address-grid!! (!! 1)))  ;; My position in the group


(*proclaim '(type Boolean-pvar act cnd shrt slvd asnd snd rcv))

(*defvar  act nil!!)        ;; Active in graph rep
(*defvar  cnd nil!!)        ;; My node is candidate for solution
(*defvar shrt nil!!)        ;; My path is minimal
(*defvar slvd nil!!)        ;; My node is solved
(*defvar asnd nil!!)        ;; I am active sender
(*defvar  snd nil!!)        ;; I am next proc to use as sender for my node
(*defvar  rcv nil!!)        ;; I am next proc to use as rec'vr for my node

;; Set up CM with virtual dimensions from mxn mxe

(defun setup ()
 (*cold-boot :initial-dimensions (list mxn mxe))
)

(defvar NODES 0)   ; Actual number of nodes
(defvar EDGES 0)   ; Actual number of edges

;; Enter edge from n to nt of length d

(defun enter-edge (n nt d)
 (let ((rc 0))
      (declare (type fixnum rc))
   (incf EDGES)
   (when (and (*or (and!! (=!! myx (!! nt)) rcv))
	      (*or (and!! (=!! myx (!! n))  snd)))
         ;; Receiver and sender OK
    (*when (and!! (=!! myx (!! nt)) rcv)
      (do-for-selected-processors (c) ; exactly one processor
        (*setf (pref act c) t)  (setq rc c)  (*setf (pref fr c) n)
 	(if (< (pref myy c) (1- mxe)) (*news rcv rcv 0 1))
        (*setf (pref rcv c) nil))
     )  ; end *when
    (*when (and!! (=!! myx (!! n)) snd)
       (do-for-selected-processors (c) ; exactly one processor
         (*setf (pref lng c) d)  (*setf (pref act c) t)
	 (*setf (pref to c) rc)  (*setf (pref asnd c) t)
	 (if (< (pref myy c) (1- mxe)) (*news snd snd 0 1))
	 (*setf (pref snd c) nil))
      )))  ; end *when, end when (and ...), end let
)  ; end enter-edge

;; Read graph description from stream and install - Cm should be setup
;; suitably.  Ignores edges which won't fit.

(defun loadg (strm)
 (*set lng (!! -1))  (*set fr (!! -1))
 (*set  act nil!!)  (*set asnd nil!!)  (*set snd nil!!)  (*set rcv nil!!)
 (*when (=!! myy (!! 0)) (*set snd t!!)  (*set rcv t!!))
 (setq NODES (setq EDGES 0))
 (do ((n (read strm) (read strm)))
     ((< n 0))
   (incf NODES)
   (do ((nt (read strm) (read strm)) (d (read strm) (read strm)))
       ((< nt 0))
     (enter-edge n nt d)
    )  ; end do nt, d
  )  ; end do n
 (list NODES EDGES)
)  ; end loadg

;; Enter node s as starting point

(defun startnode (s)
 (declare (type fixnum s))
 (*when (and!! act (=!! myx (!! s)))
     (*set cnd t!!)  (*set ptd (!! 0)))
) ; end startnode


;; Find shortest paths from node s, multiple sources if s is a vector

(defun path (s)
 (*set  cnd nil!!)  (*set shrt nil!!)  (*set slvd nil!!)
 (*set ptd (!! INF))  (*set myd (!! INF))
 (let (more (w 0) (eps 0) (k 0))
      (declare (type fixnum w eps k))
   (if (numberp s)   ;; Initialize starting node(s)
       (startnode s)
       (dotimes (k (length s)) (startnode (svref s k))))
   (*when (and!! cnd asnd) (setq more (*min lng)))
   ;; Loop terminates when there are no edges leaving candidates
   (do ()             ;; Solution loop - Booleans reflect status
       ((null more))  ;; ptd is min distance so far, or INF
     (incf k)         ;; Count iterations
     (setq eps more)  ;; Minimum length of edges leaving candidates
     (*when cnd
       (setq w (*min ptd))
       (*when (<=!! ptd (!! (+ w eps)))    ;; Newly solved nodes
         (*set slvd t!!)  (*set cnd nil!!)
	 (*when asnd (*pset :no-collisions (+!! lng ptd) myd to))
       )) ;end *when <=!!, end *when cnd
     (*when act
       (*set ptd (reduce-and-spread!! myd 'min!! 1))    ;; Min dist so far
       (*when (<!! ptd (!! (the fixnum INF)))
         (*set shrt (=!! myd ptd))
	 (*when (not!! slvd)
	   (*set cnd t!!)
	   (*when asnd (setq more (*min lng))))
       )) ; end *when <!!, *when act
   ) ; end do more
 ;; Incorporate remaining candidates into solution
 (*when cnd (*set slvd t!!)  (*set cnd nil!!))
 k) ; end let more w eps k
) ; end path

;; Print description of current graph

(defun printg ()
 (format t
    " fr   to lng   to lng   to lng   to lng   to lng   to lng   to lng")
 (*let ((more act))
       (declare (type Boolean-pvar more))
    (let ((n 0) (k 0) (nt 0) (d 0))
        (declare (type fixnum n k nt d))
     (do ()
         ((null (*when more (setq n (*min myx)))))
      (*when more
       (format t "~%~3D" n)
       (setq k 0)  ;; k is number of edges so far on the current line
       (*when (=!! myx (!! n))
         (*when asnd
          (do-for-selected-processors (c)
	     (setq nt (grid-from-cube-address (pref to c) 0))
	     (setq d  (pref lng c))
	     (when (> (incf k) 7) (format t "~%   ") (setq k 1))
	     (format t "  ~3D ~3D" nt d))) ; end do-for-, end *when plusp!!
	 (*set more nil!!)
        )  ; end *when =!!
       )))) ; end *when more, end do, end let, end *let
 ) ; end printg

;; Print all current paths

(defun printpath ()
 (format t " to  lng  from ...~%")
 (*let ((more act))
       (declare (type Boolean-pvar more))
   (let ((n 0) (k 0) (nf 0) (d 0))
         (declare (type fixnum n k nf d))
    (do ()
        ((null (*when more (setq n (*min myx)))))
     (*when more
       (setq d (min (pref ptd (grid n 0)) 9999))   ;; Shortest path length
       (format t "~%~3D ~4D" n d)
       (setq k 0)
       (*when (=!! myx (!! n))
         (*when shrt
           (do-for-selected-processors (c)
	     (setq nf (pref fr c))
	     (when (> (incf k) 16) (format t "~%        ") (setq k 0))
	     (format t " ~3D" nf)))  ; end do-for-, end *when shrt
	 (*set more nil!!)
	) ; end *when =!!
      )))) ; end *when more, end do, end let, end *let
) ; end printpath

;; Print Boolean pvar compactly according to grid

(*defun pb (bpv)
 (declare (type Boolean-pvar bpv))
 (let ((nx 0) (y 0))
  (*when act (setq nx (1+ (*max myx))))
  (dotimes (x nx)
    (terpri)
    (dotimes (y mxe) (write-char (if (pref bpv (grid x y)) #\T #\F))))
 )
)

;; Double current graph with connections of length cl.  Intended for
;; situations in which nodes are 0, ..., NODES-1 (contiguous).

(defun double (cl)
 (let ((sndn 0) (rcvn 0))
      (declare (type fixnum sndn rcvn))
   (*when snd (setq sndn (*min myx)))
   (*when rcv (setq rcvn (*min myx)))
   (when (or (null snd) (null rcv) (> (* 2 NODES) mxn))
         (format t "Can't double the graph.~%")
	 (return-from double))
   (*let ((buddy (!! 0)))
     (*if (<!! myx (!! NODES))
	  (*set buddy (cube-from-grid-address!! (+!! myx (!! NODES)) myy))
	  (*set buddy (cube-from-grid-address!! (-!! myx (!! NODES)) myy)))
     (*when (<!! myx (!! NODES))               ;; Copies everything
       (*pset :no-collisions lng lng buddy)
       (*pset :no-collisions act act buddy)    ;; Booleans sends are
       (*pset :no-collisions asnd asnd buddy)  ;; very inefficient,
       (*pset :no-collisions snd snd buddy)    ;; but why bother with
       (*pset :no-collisions rcv rcv buddy)    ;; anything else?
       (*pset :no-collisions (+!! fr (!! NODES)) fr buddy)
       (*pset :no-collisions (pref!! buddy to) to buddy))
     (setq EDGES (* 2 EDGES)) ;; Will be advanced by enter-edge
     (enter-edge sndn (+ rcvn NODES) cl)
     (enter-edge (+ sndn NODES) rcvn cl)
     (setq NODES (* 2 NODES))
   )  ; end *let
 (format t "The graph has ~D nodes, ~D edges.~%" NODES EDGES)
 )  ; end let
)  ; end double

(defvar file "gtest.dat")      ; The current data file (a string)

;; Load graph from data file whose name is in file.

(defun sg ()
  (with-open-file (strmx file) (loadg strmx))
)
